import torch
from torch import nn
from torchvision import models
import torch.nn.functional as F
from resnet import *

class AAM(nn.Module):
    def __init__(self, in_ch,out_ch):
        super(AAM, self).__init__() 
        self.global_pooling = nn.AdaptiveAvgPool1d(1)

        self.conv1 = nn.Sequential(
            nn.Conv1d(in_ch, out_ch, 1, padding=0),
            nn.BatchNorm1d(out_ch),
            nn.ReLU(inplace=True))

        self.conv2 = nn.Sequential(
            nn.Conv1d(in_ch, out_ch, 1, padding=0),
            nn.BatchNorm1d(out_ch),
            nn.ReLU(inplace=True))

        self.conv3 = nn.Sequential(
            nn.Conv1d(out_ch, out_ch, 1, padding=0),
            nn.Softmax(dim=1))

        self.conv4 = nn.Sequential(
            nn.Conv1d(in_ch, out_ch, 1, padding=0),
            nn.BatchNorm1d(out_ch),
            nn.ReLU(inplace=True))

    def forward(self, input_high, input_low):
        mid_high=self.global_pooling(input_high)
        weight_high=self.conv1(mid_high)

        mid_low = self.global_pooling(input_low)
        weight_low = self.conv2(mid_low)

        weight=self.conv3(weight_low+weight_high)
        low = self.conv4(input_low)
        return input_high+low.mul(weight)

class RAUNet(nn.Module):
    def __init__(self, num_classes=1, num_channels=3, pretrained=True):
        super().__init__()
        assert num_channels == 3
        self.w = 512
        self.h = 640
        self.num_classes = num_classes
        filters = [64, 128, 256, 512]
        resnet = ResUNet50(num_channels, num_classes=num_classes)
        #resnet = models.resnet34(pretrained=pretrained)
        # filters = [256, 512, 1024, 2048]
        # resnet = models.resnet50(pretrained=pretrained)

        self.firstconv = resnet.conv1
        self.firstbn = resnet.bn1
        self.firstrelu = resnet.relu
        self.firstmaxpool = resnet.maxpool
        self.encoder1 = resnet.layer1
        self.encoder2 = resnet.layer2
        self.encoder3 = resnet.layer3
        self.encoder4 = resnet.layer4

        # Decoder
        self.decoder4 = DecoderBlockLinkNet(filters[3], filters[2])
        self.decoder3 = DecoderBlockLinkNet(filters[2], filters[1])
        self.decoder2 = DecoderBlockLinkNet(filters[1], filters[0])
        self.decoder1 = DecoderBlockLinkNet(filters[0], filters[0])
        self.gau3 = AAM(filters[2], filters[2]) #RAUNet
        self.gau2 = AAM(filters[1], filters[1])
        self.gau1 = AAM(filters[0], filters[0])


        # Final Classifier
        self.finaldeconv1 = nn.ConvTranspose2d(filters[0], 32, 3, stride=2)
        self.finalrelu1 = nn.ReLU(inplace=True)
        self.finalconv2 = nn.Conv2d(32, 32, 3)
        self.finalrelu2 = nn.ReLU(inplace=True)
        self.finalconv3 = nn.Conv2d(32, num_classes, 2, padding=1)

    # noinspection PyCallingNonCallable
    def forward(self, x):
        # Encoder
        x = self.firstconv(x)
        x = self.firstbn(x)
        x = self.firstrelu(x)
        x = self.firstmaxpool(x)
        e1 = self.encoder1(x)
        e2 = self.encoder2(e1)
        e3 = self.encoder3(e2)
        e4 = self.encoder4(e3)

        d4 = self.decoder4(e4)
        b4 = self.gau3(d4, e3)
        d3 = self.decoder3(b4)
        b3 = self.gau2(d3, e2)
        d2 = self.decoder2(b3)
        b2 = self.gau1(d2, e1)
        d1 = self.decoder1(b2)

        # Final Classification
        f1 = self.finaldeconv1(d1)
        f2 = self.finalrelu1(f1)
        f3 = self.finalconv2(f2)
        f4 = self.finalrelu2(f3)
        f5 = self.finalconv3(f4)

        if self.num_classes > 1:
            x_out = F.log_softmax(f5, dim=1)
        else:
            x_out = f5
        return x_out
        
class DecoderBlockLinkNet(nn.Module):
    def __init__(self, in_channels, n_filters):
        super().__init__()

        self.relu = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv1d(in_channels, in_channels // 4, 1)
        self.norm1 = nn.BatchNorm1d(in_channels // 4)

        # B, C/4, H, W -> B, C/4, 2 * H, 2 * W
        self.deconv2 = nn.ConvTranspose1d(in_channels // 4, in_channels // 4, kernel_size=4, stride=2, padding=1, output_padding=0)
        self.norm2 = nn.BatchNorm1d(in_channels // 4)

        # B, C/4, H, W -> B, C, H, W
        self.conv3 = nn.Conv1d(in_channels // 4, n_filters, 1)
        self.norm3 = nn.BatchNorm1d(n_filters)


    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu(x)
        x = self.deconv2(x)
        x = self.norm2(x)
        x = self.relu(x)
        x = self.conv3(x)
        x = self.norm3(x)
        x = self.relu(x)
        return x




import os
import json

import numpy
from nn_trainer.trainers import *
from nn_trainer.trainers.abstract_nn_trainer import AbstractNnTrainer

from nn_trainer.metrics import *
from nn_trainer.callbacks import *

import torch
import torch.nn
import torch.optim
import torch.utils.data as td
from torch.autograd import grad as torch_grad


from typing import Dict

class WassersteinGanNnTrainer(AbstractNnTrainer):
    """
    Basic implementation of a generative adversarial network trainer using Wasserstein loss metric
    and backpropagation

    Args:
        AbstractNnTrainer ([type]): The base class from which this class inherits
    """
    def __init__(
        self,
        args,
        generator_neural_network: torch.nn.Module, 
        discriminator_neural_network: torch.nn.Module, 
        generator_optimizer_fn: Any = torch.optim.Adam,
        generator_optimizer_params: Dict = dict(lr=0.00005),
        discriminator_optimizer_fn: Any = torch.optim.Adam,
        discriminator_optimizer_params: Dict = dict(lr=0.00005),
        scheduler_fn: Any = None,
        scheduler_params: Dict = field(default_factory=dict),
        dtype: Any = torch.float32,
        verbose: bool = False,
        patience: int = 0,
        critic_iterations: int = 5,
        gradient_penalty_weight: int = 10,
        logger: Any = None
        ):
        """Constructor

        Args:
            args (Any): Command line args
            generator_neural_network (torch.nn.Module): [description]
            discriminator_neural_network (torch.nn.Module): [description]
            generator_optimizer_fn (Any, optional): [description]. Defaults to torch.optim.Adam.
            generator_optimizer_params (Dict, optional): [description]. Defaults to dict(lr=0.0002).
            discriminator_optimizer_fn (Any, optional): [description]. Defaults to torch.optim.Adam.
            discriminator_optimizer_params (Dict, optional): [description]. Defaults to dict(lr=0.0002).
            scheduler_fn (Any, optional): [description]. Defaults to None.
            scheduler_params (Dict, optional): [description]. Defaults to field(default_factory=dict).
            dtype (Any, optional): [description]. Defaults to torch.float32.
            verbose (bool, optional): [description]. Defaults to False.
            patience (int, optional): [description]. Defaults to 0.
            logger (Any, optional): [description]. Defaults to None.
        """
        super().__init__(
            args,
            generator_neural_network, 
            generator_optimizer_fn,
            generator_optimizer_params,
            scheduler_fn,
            scheduler_params,
            dtype,
            verbose,
            patience,
            logger)
        self._critic_iterations = critic_iterations
        self._net_d = discriminator_neural_network.type(dtype)
        self._net_d.to(self._device)
        self._net_d_optimizer_fn = discriminator_optimizer_fn
        self._net_d_optimizer_params = discriminator_optimizer_params
        self._net_d_scheduler_fn = scheduler_fn
        self._net_d_scheduler_params = scheduler_params
        self._gradient_penalty_weight = gradient_penalty_weight
        self._net_d_optimizer = self._net_d_optimizer_fn(self._net_d.parameters(), **self._net_d_optimizer_params)

        self._discriminator_model_directory_path = make_path(os.path.join(self._model_directory_path, self._net_d.__class__.__name__))

    def train(
        self,
        training_data_set: td.Dataset,
        validation_data_set: td.Dataset, 
        callbacks: List[Callback] = [], 
        loss_fn = torch.nn.MSELoss(),
        metrics: List[Metric] = [],
        ):
        """        
        The training method

        Args:
            training_data_set (td.Dataset): The dataset intended to be used for training
            validation_data_set (td.Dataset): The dataset intended to be used for validation. all metrics defined will be invokd on this set
            callbacks (List[Callback], optional): The set of callbacks receiving event based data. Defaults to [].
            loss_fn ([type], optional): The loss function to be used as the objective function during training. Defaults to torch.nn.MSELoss().
            metrics (List[Metric], optional): List of metrics to be invoked on validation data. Defaults to [].

        Raises:
            NotImplementedError: [description]
        """
        self._set_metrics(metrics, ['validation'])
        self._set_callbacks(callbacks)
        self._loss_fn = loss_fn
        
        self._dataset_train = training_data_set
        self._dataset_valid = validation_data_set
        self._train_data_loader = td.DataLoader(dataset=self._dataset_train, batch_size=self._batch_size, shuffle=True)
        self._valid_data_loader = td.DataLoader(dataset=self._dataset_valid, batch_size=self._batch_size, shuffle=True)
        
        self._eval_names = ['validation']

        with open(os.path.join(self._model_directory_path, 'training_config.json'), 'w') as f: json.dump(self._args, f, indent=4)

        self._stop_training = False
        
        
        num_steps = 0
        
        # Call method on_train_begin for all callbacks
        self._callback_container.on_train_begin()
        for epoch_index in range(self._epoch_count):
            self._callback_container.on_epoch_begin(epoch_index)
            for i, (i_real, o_real) in enumerate(self._train_data_loader):
                num_steps += 1
                batch_logs = {"batch_size": i_real.shape[0]}

                i_real = self.to_tensor(i_real, self._dtype)
                o_real = self.to_tensor(o_real, self._dtype)
                
                # Get generated data
                generated_data = self.sample_generator(i_real)

                # Calculate probabilities on real and generated data
                i_real = torch.autograd.Variable(i_real)
                o_real = torch.autograd.Variable(o_real)
                
                d_real = self._net_d(i_real, o_real)
                d_fake = self._net_d(i_real, generated_data)

                # Get gradient penalty
                gradient_penalty, grad_norm = self._gradient_penalty(i_real, o_real, generated_data)
                #self.losses['GP'].append(gradient_penalty.data[0])

                # Create total loss and optimize
                self._net_d_optimizer.zero_grad()
                d_loss = d_fake.mean() - d_real.mean() + gradient_penalty
                d_loss.backward()

                self._net_d_optimizer.step()

                # Record loss
                #batch_logs["loss"] = self.to_numpy(d_loss).item()
                batch_logs["d_loss"] = self.to_numpy(d_loss).item()

                # Only update generator every |critic_iterations| iterations
                if num_steps % self._critic_iterations == 0:
                    """ """
                    self._net_g_optimizer.zero_grad()

                    # Get generated data
                    generated_data = self.sample_generator(i_real)

                    # Calculate loss and optimize
                    d_fake = self._net_d(i_real, generated_data)
                    g_loss = -d_fake.mean()
                    g_loss.backward()   
                    self._net_g_optimizer.step()

                    # Record loss
                    batch_logs['g_loss'] = self.to_numpy(g_loss).item()
                
                self._callback_container.on_batch_end(i, batch_logs)

                # if i % self.print_every == 0:
                #     print("Iteration {}".format(i + 1))
                #     print("D: {}".format(self.losses['D'][-1]))
                #     print("GP: {}".format(self.losses['GP'][-1]))
                #     print("Gradient norm: {}".format(self.losses['gradient_norm'][-1]))
                #     if self.num_steps > self.critic_iterations:
                #         print("G: {}".format(self.losses['G'][-1]))
            epoch_logs = { 
                "lr_net_g": self._net_g_optimizer.param_groups[-1]["lr"],
                "lr_net_d": self._net_d_optimizer.param_groups[-1]["lr"] 
            }
            self.history.epoch_metrics.update(epoch_logs)

            # Apply predict epoch to all validation set
            for eval_name, valid_data in zip(self._eval_names, self._valid_data_loader):
                self._predict_epoch(eval_name, valid_data)

            # Call method on_epoch_end for all callbacks
            self._callback_container.on_epoch_end(epoch_index, logs=self.history.epoch_metrics)
            if self._stop_training:
                break
        
        # Call method on_train_end for all callbacks
        self._callback_container.on_train_end()
        self._net_g.eval()
        self._net_d.eval()
    
    def sample_generator(self, inputs):
        generated_data = self._net_g(inputs)
        return generated_data

    def _gradient_penalty(self, real_inputs, real_data, gen_data):
        batch_size = real_data.size()[0]
        dims = numpy.ones(len(real_data.shape), dtype=numpy.int64)
        dims[0] = batch_size
        dims = tuple(dims)
        t = torch.rand(dims, requires_grad=True)
        t = t.expand_as(real_data)

        if self._is_cuda_enabled:
            t = t.cuda()

        # mixed sample from real and fake; make approx of the 'true' gradient norm
        interpol = t * real_data.data + (1-t) * gen_data.data

        if self._is_cuda_enabled:
            interpol = interpol.cuda()

        #interpol = torch.tensor(interpol, requires_grad=True)
        
        prob_interpol = self._net_d(real_inputs, interpol)
        torch.autograd.set_detect_anomaly(True)
        gradients = torch_grad(outputs=prob_interpol, inputs=interpol, grad_outputs=torch.ones(prob_interpol.size()).cuda() if self._is_cuda_enabled else torch.ones(prob_interpol.size()), create_graph=True, retain_graph=True)[0]
        gradients = gradients.view(batch_size, -1)
        #grad_norm = torch.norm(gradients, dim=1).mean()
        #self.losses['gradient_norm'].append(grad_norm.item())

        # add epsilon for stability
        eps = 1e-10
        gradients_norm = torch.sqrt(torch.sum(gradients**2, dim=1, dtype=torch.double) + eps)
        #gradients = gradients.cpu()
        # comment: precision is lower than grad_norm (think that is double) and gradients_norm is float
        return self._gradient_penalty_weight * (torch.max(torch.zeros(1,dtype=torch.double).cuda() if self._is_cuda_enabled else torch.zeros(1,dtype=torch.double), gradients_norm.mean() - 1) ** 2), gradients_norm.mean().item()

    def _set_callbacks(self, custom_callbacks):
        """Setup the callbacks functions.
        Parameters
        ----------
        custom_callbacks : list of func
            List of callback functions.
        """
        # Setup default callbacks history, early stopping and scheduler
        callbacks = []
        self.history = History(self, verbose=self._verbose, logger=self._logger)
        callbacks.append(self.history)
        if (self.early_stopping_metric is not None) and (self._patience > 0):
            early_stopping = EarlyStopping(early_stopping_metric=self.early_stopping_metric, is_maximize=(self._metrics[-1]._maximize if len(self._metrics) > 0 else None), patience=self._patience, logger=self._logger)
            callbacks.append(early_stopping)
        else:
            if self._logger is not None:
                self._logger.info("No early stopping will be performed, last training weights will be used.")
        if self._net_g_scheduler_fn is not None:
            # Add LR Scheduler call_back for generator
            is_batch_level = self._net_g_scheduler_params.pop("is_batch_level", False)
            net_g_scheduler = LRSchedulerCallback(scheduler_fn=self._net_g_scheduler_fn, scheduler_params=self._net_g_scheduler_params, optimizer=self._net_g_optimizer, early_stopping_metric=self.early_stopping_metric, is_batch_level=is_batch_level)
            callbacks.append(net_g_scheduler)
        if self._net_d_scheduler_fn is not None:
            # Add LR Scheduler call_back for discriminator
            is_batch_level = self._net_d_scheduler_params.pop("is_batch_level", False)
            net_d_scheduler = LRSchedulerCallback(scheduler_fn=self._net_d_scheduler_fn, scheduler_params=self._net_d_scheduler_params, optimizer=self._net_d_optimizer, early_stopping_metric=self.early_stopping_metric, is_batch_level=is_batch_level)
            callbacks.append(net_d_scheduler)

        if custom_callbacks:
            callbacks.extend(custom_callbacks)
        self._callback_container = CallbackContainer(callbacks)
        self._callback_container.set_trainer(self)


import os
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from nn_trainer.plotters import DataPlotter
matplotlib.use('Agg')

def make_path(output_path):
    if not os.path.isdir(output_path):
        os.makedirs(output_path)
    return output_path

class ScatterDataPlotter(DataPlotter):
    def plot_batch(self, train_set_inputs, train_set_output, synthetic_output, epoch, output_dir):
        """
        Save output samples.
        """
        if len(train_set_inputs.shape) == 2 and len(train_set_output.shape) == 2 and len(synthetic_output.shape) == 2:
            train_set_inputs = train_set_inputs.reshape(train_set_inputs.shape[0], 1, train_set_inputs.shape[1])
            train_set_output = train_set_output.reshape(train_set_output.shape[0], 1, train_set_output.shape[1])
            synthetic_output = synthetic_output.reshape(synthetic_output.shape[0], 1, synthetic_output.shape[1])

        assert synthetic_output.shape == train_set_output.shape
        assert len(train_set_inputs.shape) == 3 and len(train_set_output.shape) == 3 and len(synthetic_output.shape) == 3

        plt.clf()
        sample_dir = make_path(os.path.join(output_dir, str(epoch)))
        width = 0.4  # width of bar

        for idx, sample in enumerate(synthetic_output):
            output_path = os.path.join(sample_dir, "{}_noise".format(idx + 1))

            sample = sample[0]
            x_output = np.arange(len(sample))
            nrows = train_set_inputs.shape[1] + train_set_output.shape[1]
            plt.figure(figsize=(20, 10))
            for ti in range(train_set_inputs.shape[1]):
                train_input = train_set_inputs[idx][ti]
                x_input = np.arange(len(train_input))
                ax = plt.subplot(nrows, 1, ti + 1)
                ax.set_ylim([-1, 1])
                #ax.set_xticks(x_input)

                plt.scatter(x_input, train_input, label='Input_{}'.format(ti), color='indigo')
                plt.ylabel('Input {}'.format(ti))

                plt.ylim(-1.1, 1.1)
                plt.legend(loc=4)
                plt.grid(axis='x')

            train_set_sample = train_set_output[idx][0]
            # share x and y
            ax3 = plt.subplot(nrows, 1, nrows)
            ax3.set_ylim([-1, 1])
            #ax3.set_xticks(x_output)

            plt.scatter(x_output, sample, label='G_output', color='dodgerblue')
            plt.scatter(x_output, train_set_sample, label='Validation_Output', color='mediumvioletred')

            plt.xlabel('Sample Index')
            plt.ylabel('G Output')

            plt.legend(loc=4)
            #plt.grid(axis='x')
            plt.tight_layout()

            plt.savefig(output_path + '.png')
            plt.clf()
            plt.close()



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init

def init_weights(net, init_type='normal', gain=0.02):
    def init_func(m):
        classname = m.__class__.__name__
        if hasattr(m, 'weight') and (classname.find('Conv') != -1 or classname.find('Linear') != -1):
            if init_type == 'normal':
                init.normal_(m.weight.data, 0.0, gain)
            elif init_type == 'xavier':
                init.xavier_normal_(m.weight.data, gain=gain)
            elif init_type == 'kaiming':
                init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                init.orthogonal_(m.weight.data, gain=gain)
            else:
                raise NotImplementedError('initialization method [%s] is not implemented' % init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                init.constant_(m.bias.data, 0.0)
        elif classname.find('BatchNorm1d') != -1:
            init.normal_(m.weight.data, 1.0, gain)
            init.constant_(m.bias.data, 0.0)

    print('initialize network with %s' % init_type)
    net.apply(init_func)

class conv_block(nn.Module):
    def __init__(self,ch_in,ch_out):
        super(conv_block,self).__init__()
        self.conv = nn.Sequential(
            nn.Conv1d(ch_in, ch_out, kernel_size=3,stride=1,padding=1,bias=True),
            nn.BatchNorm1d(ch_out),
            nn.ReLU(inplace=True),
            nn.Conv1d(ch_out, ch_out, kernel_size=3,stride=1,padding=1,bias=True),
            nn.BatchNorm1d(ch_out),
            nn.ReLU(inplace=True)
        )


    def forward(self,x):
        x = self.conv(x)
        return x

class up_conv(nn.Module):
    def __init__(self,ch_in,ch_out):
        super(up_conv,self).__init__()
        self.up = nn.Sequential(
            nn.Upsample(scale_factor=2),
            nn.Conv1d(ch_in,ch_out,kernel_size=3,stride=1,padding=1,bias=True),
		    nn.BatchNorm1d(ch_out),
			nn.ReLU(inplace=True)
        )

    def forward(self,x):
        x = self.up(x)
        return x

class Recurrent_block(nn.Module):
    def __init__(self,ch_out,t=2):
        super(Recurrent_block,self).__init__()
        self.t = t
        self.ch_out = ch_out
        self.conv = nn.Sequential(
            nn.Conv1d(ch_out,ch_out,kernel_size=3,stride=1,padding=1,bias=True),
		    nn.BatchNorm1d(ch_out),
			nn.ReLU(inplace=True)
        )

    def forward(self,x):
        for i in range(self.t):

            if i==0:
                x1 = self.conv(x)
            
            x1 = self.conv(x+x1)
        return x1
        
class RRCNN_block(nn.Module):
    def __init__(self,ch_in,ch_out,t=2):
        super(RRCNN_block,self).__init__()
        self.RCNN = nn.Sequential(
            Recurrent_block(ch_out,t=t),
            Recurrent_block(ch_out,t=t)
        )
        self.Conv_1x1 = nn.Conv1d(ch_in,ch_out,kernel_size=1,stride=1,padding=0)

    def forward(self,x):
        x = self.Conv_1x1(x)
        x1 = self.RCNN(x)
        return x+x1


class single_conv(nn.Module):
    def __init__(self,ch_in,ch_out):
        super(single_conv,self).__init__()
        self.conv = nn.Sequential(
            nn.Conv1d(ch_in, ch_out, kernel_size=3,stride=1,padding=1,bias=True),
            nn.BatchNorm1d(ch_out),
            nn.ReLU(inplace=True)
        )

    def forward(self,x):
        x = self.conv(x)
        return x

class Attention_block(nn.Module):
    def __init__(self,F_g,F_l,F_int):
        super(Attention_block,self).__init__()
        self.W_g = nn.Sequential(
            nn.Conv1d(F_g, F_int, kernel_size=1,stride=1,padding=0,bias=True),
            nn.BatchNorm1d(F_int)
            )
        
        self.W_x = nn.Sequential(
            nn.Conv1d(F_l, F_int, kernel_size=1,stride=1,padding=0,bias=True),
            nn.BatchNorm1d(F_int)
        )

        self.psi = nn.Sequential(
            nn.Conv1d(F_int, 1, kernel_size=1,stride=1,padding=0,bias=True),
            nn.BatchNorm1d(1),
            nn.Sigmoid()
        )
        
        self.relu = nn.ReLU(inplace=True)
        
    def forward(self,g,x):
        g1 = self.W_g(g)
        x1 = self.W_x(x)
        psi = self.relu(g1+x1)
        psi = self.psi(psi)

        return x*psi


class U_Net(nn.Module):
    def __init__(self,img_ch=3,output_ch=1):
        super(U_Net,self).__init__()
        
        self.Maxpool = nn.MaxPool1d(kernel_size=2,stride=2)

        self.Conv1 = conv_block(ch_in=img_ch,ch_out=64)
        self.Conv2 = conv_block(ch_in=64,ch_out=128)
        self.Conv3 = conv_block(ch_in=128,ch_out=256)
        self.Conv4 = conv_block(ch_in=256,ch_out=512)
        self.Conv5 = conv_block(ch_in=512,ch_out=1024)

        self.Up5 = up_conv(ch_in=1024,ch_out=512)
        self.Up_conv5 = conv_block(ch_in=1024, ch_out=512)

        self.Up4 = up_conv(ch_in=512,ch_out=256)
        self.Up_conv4 = conv_block(ch_in=512, ch_out=256)
        
        self.Up3 = up_conv(ch_in=256,ch_out=128)
        self.Up_conv3 = conv_block(ch_in=256, ch_out=128)
        
        self.Up2 = up_conv(ch_in=128,ch_out=64)
        self.Up_conv2 = conv_block(ch_in=128, ch_out=64)

        self.Conv_1x1 = nn.Conv1d(64,output_ch,kernel_size=1,stride=1,padding=0)

    def forward(self,x):
        # encoding path
        x1 = self.Conv1(x)

        x2 = self.Maxpool(x1)
        x2 = self.Conv2(x2)
        
        x3 = self.Maxpool(x2)
        x3 = self.Conv3(x3)

        x4 = self.Maxpool(x3)
        x4 = self.Conv4(x4)

        x5 = self.Maxpool(x4)
        x5 = self.Conv5(x5)

        # decoding + concat path
        d5 = self.Up5(x5)
        d5 = torch.cat((x4, d5), dim=1)
        
        d5 = self.Up_conv5(d5)
        
        d4 = self.Up4(d5)
        d4 = torch.cat((x3, d4), dim=1)
        d4 = self.Up_conv4(d4)

        d3 = self.Up3(d4)
        d3 = torch.cat((x2, d3), dim=1)
        d3 = self.Up_conv3(d3)

        d2 = self.Up2(d3)
        d2 = torch.cat((x1, d2), dim=1)
        d2 = self.Up_conv2(d2)

        d1 = self.Conv_1x1(d2)

        return torch.tanh(d1)
        #return d1


class R2U_Net(nn.Module):
    def __init__(self,img_ch=3,output_ch=1,t=2):
        super(R2U_Net,self).__init__()
        
        self.Maxpool = nn.MaxPool1d(kernel_size=2,stride=2)
        self.Upsample = nn.Upsample(scale_factor=2)

        self.RRCNN1 = RRCNN_block(ch_in=img_ch,ch_out=64,t=t)

        self.RRCNN2 = RRCNN_block(ch_in=64,ch_out=128,t=t)
        
        self.RRCNN3 = RRCNN_block(ch_in=128,ch_out=256,t=t)
        
        self.RRCNN4 = RRCNN_block(ch_in=256,ch_out=512,t=t)
        
        self.RRCNN5 = RRCNN_block(ch_in=512,ch_out=1024,t=t)
        

        self.Up5 = up_conv(ch_in=1024,ch_out=512)
        self.Up_RRCNN5 = RRCNN_block(ch_in=1024, ch_out=512,t=t)
        
        self.Up4 = up_conv(ch_in=512,ch_out=256)
        self.Up_RRCNN4 = RRCNN_block(ch_in=512, ch_out=256,t=t)
        
        self.Up3 = up_conv(ch_in=256,ch_out=128)
        self.Up_RRCNN3 = RRCNN_block(ch_in=256, ch_out=128,t=t)
        
        self.Up2 = up_conv(ch_in=128,ch_out=64)
        self.Up_RRCNN2 = RRCNN_block(ch_in=128, ch_out=64,t=t)

        self.Conv_1x1 = nn.Conv1d(64,output_ch,kernel_size=1,stride=1,padding=0)


    def forward(self,x):
        # encoding path
        x1 = self.RRCNN1(x)

        x2 = self.Maxpool(x1)
        x2 = self.RRCNN2(x2)
        
        x3 = self.Maxpool(x2)
        x3 = self.RRCNN3(x3)

        x4 = self.Maxpool(x3)
        x4 = self.RRCNN4(x4)

        x5 = self.Maxpool(x4)
        x5 = self.RRCNN5(x5)

        # decoding + concat path
        d5 = self.Up5(x5)
        d5 = torch.cat((x4,d5),dim=1)
        d5 = self.Up_RRCNN5(d5)
        
        d4 = self.Up4(d5)
        d4 = torch.cat((x3,d4),dim=1)
        d4 = self.Up_RRCNN4(d4)

        d3 = self.Up3(d4)
        d3 = torch.cat((x2,d3),dim=1)
        d3 = self.Up_RRCNN3(d3)

        d2 = self.Up2(d3)
        d2 = torch.cat((x1,d2),dim=1)
        d2 = self.Up_RRCNN2(d2)

        d1 = self.Conv_1x1(d2)

        return torch.tanh(d1)
        #return d1


class AttU_Net(nn.Module):
    def __init__(self,img_ch=3,output_ch=1):
        super(AttU_Net,self).__init__()
        
        self.Maxpool = nn.MaxPool1d(kernel_size=2,stride=2)

        self.Conv1 = conv_block(ch_in=img_ch,ch_out=64)
        self.Conv2 = conv_block(ch_in=64,ch_out=128)
        self.Conv3 = conv_block(ch_in=128,ch_out=256)
        self.Conv4 = conv_block(ch_in=256,ch_out=512)
        self.Conv5 = conv_block(ch_in=512,ch_out=1024)

        self.Up5 = up_conv(ch_in=1024,ch_out=512)
        self.Att5 = Attention_block(F_g=512,F_l=512,F_int=256)
        self.Up_conv5 = conv_block(ch_in=1024, ch_out=512)

        self.Up4 = up_conv(ch_in=512,ch_out=256)
        self.Att4 = Attention_block(F_g=256,F_l=256,F_int=128)
        self.Up_conv4 = conv_block(ch_in=512, ch_out=256)
        
        self.Up3 = up_conv(ch_in=256,ch_out=128)
        self.Att3 = Attention_block(F_g=128,F_l=128,F_int=64)
        self.Up_conv3 = conv_block(ch_in=256, ch_out=128)
        
        self.Up2 = up_conv(ch_in=128,ch_out=64)
        self.Att2 = Attention_block(F_g=64,F_l=64,F_int=32)
        self.Up_conv2 = conv_block(ch_in=128, ch_out=64)

        self.Conv_1x1 = nn.Conv1d(64,output_ch,kernel_size=1,stride=1,padding=0)


    def forward(self,x):
        # encoding path
        x1 = self.Conv1(x)

        x2 = self.Maxpool(x1)
        x2 = self.Conv2(x2)
        
        x3 = self.Maxpool(x2)
        x3 = self.Conv3(x3)

        x4 = self.Maxpool(x3)
        x4 = self.Conv4(x4)

        x5 = self.Maxpool(x4)
        x5 = self.Conv5(x5)

        # decoding + concat path
        d5 = self.Up5(x5)
        x4 = self.Att5(g=d5,x=x4)
        d5 = torch.cat((x4,d5),dim=1)        
        d5 = self.Up_conv5(d5)
        
        d4 = self.Up4(d5)
        x3 = self.Att4(g=d4,x=x3)
        d4 = torch.cat((x3,d4),dim=1)
        d4 = self.Up_conv4(d4)

        d3 = self.Up3(d4)
        x2 = self.Att3(g=d3,x=x2)
        d3 = torch.cat((x2,d3),dim=1)
        d3 = self.Up_conv3(d3)

        d2 = self.Up2(d3)
        x1 = self.Att2(g=d2,x=x1)
        d2 = torch.cat((x1,d2),dim=1)
        d2 = self.Up_conv2(d2)

        d1 = self.Conv_1x1(d2)

        return torch.tanh(d1)
        #return d1

class R2AttU_Net(nn.Module):
    def __init__(self,img_ch=3,output_ch=1,t=2):
        super(R2AttU_Net,self).__init__()
        
        self.Maxpool = nn.MaxPool1d(kernel_size=2,stride=2)
        self.Upsample = nn.Upsample(scale_factor=2)

        self.RRCNN1 = RRCNN_block(ch_in=img_ch,ch_out=64,t=t)

        self.RRCNN2 = RRCNN_block(ch_in=64,ch_out=128,t=t)
        
        self.RRCNN3 = RRCNN_block(ch_in=128,ch_out=256,t=t)
        
        self.RRCNN4 = RRCNN_block(ch_in=256,ch_out=512,t=t)
        
        self.RRCNN5 = RRCNN_block(ch_in=512,ch_out=1024,t=t)
        

        self.Up5 = up_conv(ch_in=1024,ch_out=512)
        self.Att5 = Attention_block(F_g=512,F_l=512,F_int=256)
        self.Up_RRCNN5 = RRCNN_block(ch_in=1024, ch_out=512,t=t)
        
        self.Up4 = up_conv(ch_in=512,ch_out=256)
        self.Att4 = Attention_block(F_g=256,F_l=256,F_int=128)
        self.Up_RRCNN4 = RRCNN_block(ch_in=512, ch_out=256,t=t)
        
        self.Up3 = up_conv(ch_in=256,ch_out=128)
        self.Att3 = Attention_block(F_g=128,F_l=128,F_int=64)
        self.Up_RRCNN3 = RRCNN_block(ch_in=256, ch_out=128,t=t)
        
        self.Up2 = up_conv(ch_in=128,ch_out=64)
        self.Att2 = Attention_block(F_g=64,F_l=64,F_int=32)
        self.Up_RRCNN2 = RRCNN_block(ch_in=128, ch_out=64,t=t)

        self.Conv_1x1 = nn.Conv1d(64,output_ch,kernel_size=1,stride=1,padding=0)


    def forward(self,x):
        # encoding path
        x1 = self.RRCNN1(x)

        x2 = self.Maxpool(x1)
        x2 = self.RRCNN2(x2)
        
        x3 = self.Maxpool(x2)
        x3 = self.RRCNN3(x3)

        x4 = self.Maxpool(x3)
        x4 = self.RRCNN4(x4)

        x5 = self.Maxpool(x4)
        x5 = self.RRCNN5(x5)

        # decoding + concat path
        d5 = self.Up5(x5)
        x4 = self.Att5(g=d5,x=x4)
        d5 = torch.cat((x4,d5),dim=1)
        d5 = self.Up_RRCNN5(d5)
        
        d4 = self.Up4(d5)
        x3 = self.Att4(g=d4,x=x3)
        d4 = torch.cat((x3,d4),dim=1)
        d4 = self.Up_RRCNN4(d4)

        d3 = self.Up3(d4)
        x2 = self.Att3(g=d3,x=x2)
        d3 = torch.cat((x2,d3),dim=1)
        d3 = self.Up_RRCNN3(d3)

        d2 = self.Up2(d3)
        x1 = self.Att2(g=d2,x=x1)
        d2 = torch.cat((x1,d2),dim=1)
        d2 = self.Up_RRCNN2(d2)

        d1 = self.Conv_1x1(d2)

        return torch.tanh(d1)
        #return d1


import os
import time
import math
import logging
import argparse
import numpy as np

from scipy import interpolate

def init_console_logger(logger, verbose=False):
    stream_handler = logging.StreamHandler()
    if verbose:
        stream_handler.setLevel(logging.DEBUG)
    else:
        stream_handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(levelname)s] %(message)s')
    stream_handler.setFormatter(formatter)
    file_handler = logging.FileHandler("model.log")
    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)

# Training
EPOCHS = 150
BATCH_SIZE = 128

SAMPLE_EVERY = 10  # Generate audio samples every 1 epoch.
SAMPLE_NUM = 10  # Generate 10 samples every sample generation.

OUTPUT_PATH = os.path.join(os.getcwd(), "output")

# =============Logger===============
LOGGER = logging.getLogger('nn_logger')
LOGGER.setLevel(logging.DEBUG)

LOGGER.info('Initialized logger.')
init_console_logger(LOGGER)

def make_path(output_path):
    if not os.path.isdir(output_path):
        os.makedirs(output_path)
    return output_path

def time_since(since):
    now = time.time()
    s = now - since
    m = math.floor(s / 60)
    s -= m * 60
    return '%dm %ds' % (m, s)

output = make_path(OUTPUT_PATH)

def parse_arguments():
    """
    Get command line arguments
    """
    parser = argparse.ArgumentParser(description='Train a 1d GAN with 2 input arrays on a given set 1d arrays')
    parser.add_argument('-lra', '--lrelu-alpha', dest='alpha', type=float, default=0.2, help='Slope of negative part of LReLU used by discriminator')
    parser.add_argument('-bs', '--batch-size', dest='batch_size', type=int, default=BATCH_SIZE, help='Batch size used for training')
    parser.add_argument('-ne', '--max-epoch-count', dest='max_epoch_count', type=int, default=EPOCHS, help='Number of epochs')
    parser.add_argument('-ld', '--latent-size', dest='latent_size', type=int, default=0, help='Size of latent dimension used by generator')
    parser.add_argument('-eps', '--epochs-per-sample', dest='epochs_per_sample', type=int, default=SAMPLE_EVERY, help='How many epochs between every set of samples generated for inspection')
    parser.add_argument('-ss', '--sample-size', dest='sample_size', type=int, default=SAMPLE_NUM, help='Number of inspection samples generated')
    parser.add_argument('-rf', '--regularization-factor', dest='lmbda', type=float, default=10.0, help='Gradient penalty regularization factor')
    parser.add_argument('-lr', '--learning-rate', dest='learning_rate', type=float, default=0.0001, help='Initial ADAM learning rate')
    parser.add_argument('-bo', '--beta-one', dest='beta1', type=float, default=0.5, help='beta_1 ADAM parameter')
    parser.add_argument('-bt', '--beta-two', dest='beta2', type=float, default=0.9, help='beta_2 ADAM parameter')
    parser.add_argument('-v', '--verbose', dest='verbose', default=True, action='store_true')
    parser.add_argument('-od', '--output-dir', dest='output_dir', type=str, default=output, help='Path to directory where model files will be output')
    parser.add_argument('-tr', '--testing-ratio', dest='testing_ratio', type=int, default=1, help='The subet ratio of the data set to test on')
    parser.add_argument('-vr', '--validation-ratio', dest='validation_ratio', type=int, default=1, help='The subet ratio of the data set to validate on')
    parser.add_argument('-nsm', '--save-models-count', dest='save_models_count', type=int, default=15, help='The number of epoch based models to save from the last epoch')
    args = parser.parse_args()
    return vars(args)

def interpolate_1D(array, new_array_size):
    array_size = len(array)
    x = np.arange(0, array_size)
    y = array
    interpolation_function = interpolate.interp1d(x, y, fill_value="extrapolate")
    interpolation_ratio = array_size / new_array_size
    xnew = np.arange(0, array_size, interpolation_ratio)
    ynew = interpolation_function(xnew)
    return ynew

def interpolate_2D(array, new_array_shape):
    x_sz = array.shape[0]
    y_sz = array.shape[1]
    x = np.arange(0, x_sz)
    y = np.arange(0, y_sz)
    z = array
    interpolation_function = interpolate.interp2d(y, x, z)
    x_interpolation_ratio = x_sz / new_array_shape[0]
    y_interpolation_ratio = y_sz / new_array_shape[1]
    xnew = np.arange(0, x_sz, x_interpolation_ratio)
    ynew = np.arange(0, y_sz, y_interpolation_ratio)
    znew = interpolation_function(ynew, xnew)
    return znew



import os
import json

import numpy
from nn_trainer.trainers.abstract_nn_trainer import AbstractNnTrainer


from nn_trainer.metrics import *
from nn_trainer.callbacks import *
from nn_trainer.trainers import *

import torch
import torch.nn
import torch.optim
import torch.utils.data as td

from typing import Dict

class VanillaGanNnTrainer(AbstractNnTrainer):
    """
    Basic implementation of a generative adversarial network trainer using custom loss metrics
    and backpropagation

    Args:
        AbstractNnTrainer ([type]): The base class from which this class inherits
    """
    def __init__(
        self,
        args,
        generator_neural_network: torch.nn.Module, 
        discriminator_neural_network: torch.nn.Module, 
        generator_optimizer_fn: Any = torch.optim.Adam,
        generator_optimizer_params: Dict = dict(lr=0.0002),
        discriminator_optimizer_fn: Any = torch.optim.Adam,
        discriminator_optimizer_params: Dict = dict(lr=0.0002),
        scheduler_fn: Any = None,
        scheduler_params: Dict = field(default_factory=dict),
        dtype: Any = torch.float32,
        verbose: bool = False,
        patience: int = 0,
        logger: Any = None,
        ):
        """[summary]

        Args:
            args (Any): Command line args
            generator_neural_network (torch.nn.Module): [description]
            discriminator_neural_network (torch.nn.Module): [description]
            generator_optimizer_fn (Any, optional): [description]. Defaults to torch.optim.Adam.
            generator_optimizer_params (Dict, optional): [description]. Defaults to dict(lr=0.0002).
            discriminator_optimizer_fn (Any, optional): [description]. Defaults to torch.optim.Adam.
            discriminator_optimizer_params (Dict, optional): [description]. Defaults to dict(lr=0.0002).
            scheduler_fn (Any, optional): [description]. Defaults to None.
            scheduler_params (Dict, optional): [description]. Defaults to field(default_factory=dict).
            dtype (Any, optional): [description]. Defaults to torch.float32.
            verbose (bool, optional): [description]. Defaults to False.
            patience (int, optional): [description]. Defaults to 0.
            logger (Any, optional): [description]. Defaults to None.
        """
        super().__init__(
            args,
            generator_neural_network, 
            generator_optimizer_fn,
            generator_optimizer_params,
            scheduler_fn,
            scheduler_params,
            dtype,
            verbose,
            patience,
            logger)
        self._net_d = discriminator_neural_network.type(dtype)
        self._net_d.to(self._device)
        self._net_d_optimizer_fn = discriminator_optimizer_fn
        self._net_d_optimizer_params = discriminator_optimizer_params
        self._net_d_scheduler_fn = scheduler_fn
        self._net_d_scheduler_params = scheduler_params

        self._net_d_optimizer = self._net_d_optimizer_fn(self._net_d.parameters(), **self._net_d_optimizer_params)

        self._discriminator_model_directory_path = make_path(os.path.join(self._model_directory_path, self._net_d.__class__.__name__))

    def train(
        self,
        training_data_set: td.Dataset,
        validation_data_set: td.Dataset, 
        callbacks: List[Callback] = [], 
        loss_fn = torch.nn.BCELoss(),
        metrics: List[Metric] = [],
        ):
        """        
        The training method

        Args:
            training_data_set (td.Dataset): The dataset intended to be used for training
            validation_data_set (td.Dataset): The dataset intended to be used for validation. all metrics defined will be invokd on this set
            callbacks (List[Callback], optional): The set of callbacks receiving event based data. Defaults to [].
            loss_fn ([type], optional): The loss function to be used as the objective function during training. Defaults to torch.nn.MSELoss().
            metrics (List[Metric], optional): List of metrics to be invoked on validation data. Defaults to [].

        Raises:
            NotImplementedError: [description]
        """
        self._set_metrics(metrics, ['validation'])
        self._set_callbacks(callbacks)
        self._loss_fn = loss_fn
        
        self._dataset_train = training_data_set
        self._dataset_valid = validation_data_set
        self._train_data_loader = td.DataLoader(dataset=self._dataset_train, batch_size=self._batch_size, shuffle=True)
        self._valid_data_loader = td.DataLoader(dataset=self._dataset_valid, batch_size=self._batch_size, shuffle=True)
        
        self._eval_names = ['validation']

        with open(os.path.join(self._model_directory_path, 'training_config.json'), 'w') as f: json.dump(self._args, f, indent=4)

        self._stop_training = False

        # Call method on_train_begin for all callbacks
        self._callback_container.on_train_begin()
        
        for epoch_index in range(self._epoch_count):
            self._callback_container.on_epoch_begin(epoch_index)
            self._net_g.train()
            self._net_d.train()
            
            # train on batches here
            for batch_index, (i_real, o_real) in enumerate(self._train_data_loader):
                self._callback_container.on_batch_begin(batch_index)

                batch_logs = {"batch_size": i_real.shape[0]}

                i_real = self.to_tensor(i_real, self._dtype)
                o_real = self.to_tensor(o_real, self._dtype)

                real = self.to_tensor(numpy.ones(shape=(o_real.shape[0], 1)), self._dtype)
                fake = self.to_tensor(numpy.zeros(shape=(o_real.shape[0], 1)), self._dtype)
                
                ######################
                ## Train generator
                ######################

                # zero_grad here
                self._net_g_optimizer.zero_grad()
                #for param in self._net_g.parameters(): param.grad = None

                o_fake = self._net_g(i_real)
                d_result = self._net_d(i_real, o_fake)

                g_loss = self._loss_fn(d_result, real)

                # Perform backward pass and optimization    
                g_loss.backward()
                torch.nn.utils.clip_grad_norm_(self._net_g.parameters(), 1)
                self._net_g_optimizer.step()

                batch_logs["g_loss"] = self.to_numpy(g_loss).item()
                
                ######################
                ## Train discriminator
                ######################

                # zero_grad here
                self._net_d_optimizer.zero_grad()
                #for param in self._net_d.parameters(): param.grad = None

                # Measure discriminator's ability to classify real from generated samples
                real_loss = self._loss_fn(self._net_d(i_real, o_real), real)
                fake_loss = self._loss_fn(self._net_d(i_real, o_fake.detach()), fake)
                d_loss = (real_loss + fake_loss) / 2

                d_loss.backward()
                self._net_d_optimizer.step()

                batch_logs["loss"] = self.to_numpy(d_loss).item()
                batch_logs["d_loss"] = self.to_numpy(d_loss).item()

                self._callback_container.on_batch_end(batch_index, batch_logs)
            
            epoch_logs = { 
                "lr_net_g": self._net_g_optimizer.param_groups[-1]["lr"],
                "lr_net_d": self._net_d_optimizer.param_groups[-1]["lr"] 
                }
            self.history.epoch_metrics.update(epoch_logs)

            # Apply predict epoch to all validation set
            for eval_name, valid_data in zip(self._eval_names, self._valid_data_loader):
                self._predict_epoch(eval_name, valid_data)

            # Call method on_epoch_end for all callbacks
            self._callback_container.on_epoch_end(epoch_index, logs=self.history.epoch_metrics)

            if self._stop_training:
                break
        
        # Call method on_train_end for all callbacks
        self._callback_container.on_train_end()
        self._net_g.eval()
        self._net_d.eval()
       
    def _set_callbacks(self, custom_callbacks):
        """Setup the callbacks functions.
        Parameters
        ----------
        custom_callbacks : list of func
            List of callback functions.
        """
        # Setup default callbacks history, early stopping and scheduler
        callbacks = []
        self.history = History(self, verbose=self._verbose, logger=self._logger)
        callbacks.append(self.history)
        if (self.early_stopping_metric is not None) and (self._patience > 0):
            early_stopping = EarlyStopping(early_stopping_metric=self.early_stopping_metric, is_maximize=(self._metrics[-1]._maximize if len(self._metrics) > 0 else None), patience=self._patience, logger=self._logger)
            callbacks.append(early_stopping)
        else:
            if self._logger is not None:
                self._logger.info("No early stopping will be performed, last training weights will be used.")
        if self._net_g_scheduler_fn is not None:
            # Add LR Scheduler call_back for generator
            is_batch_level = self._net_g_scheduler_params.pop("is_batch_level", False)
            net_g_scheduler = LRSchedulerCallback(scheduler_fn=self._net_g_scheduler_fn, scheduler_params=self._net_g_scheduler_params, optimizer=self._net_g_optimizer, early_stopping_metric=self.early_stopping_metric, is_batch_level=is_batch_level)
            callbacks.append(net_g_scheduler)
        if self._net_d_scheduler_fn is not None:
            # Add LR Scheduler call_back for discriminator
            is_batch_level = self._net_d_scheduler_params.pop("is_batch_level", False)
            net_d_scheduler = LRSchedulerCallback(scheduler_fn=self._net_d_scheduler_fn, scheduler_params=self._net_d_scheduler_params, optimizer=self._net_d_optimizer, early_stopping_metric=self.early_stopping_metric, is_batch_level=is_batch_level)
            callbacks.append(net_d_scheduler)

        if custom_callbacks:
            callbacks.extend(custom_callbacks)
        self._callback_container = CallbackContainer(callbacks)
        self._callback_container.set_trainer(self)


import os
import time
import datetime
import copy
import random

import numpy as np
from dataclasses import dataclass, field
from typing import Dict, List, Any
import torch
import torch.nn
import torch.utils.data
import matplotlib.pyplot as plt

from nn_trainer.plotters import DataPlotter

class Callback:
    """
    Abstract base class used to build new callbacks.
    """

    def __init__(self):
        pass

    def set_params(self, params):
        self.params = params

    def set_trainer(self, model):
        self.trainer = model

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        pass

    def on_batch_begin(self, batch, logs=None):
        pass

    def on_batch_end(self, batch, logs=None):
        pass

    def on_train_begin(self, logs=None):
        pass

    def on_train_end(self, logs=None):
        pass

@dataclass
class CallbackContainer:
    """
    Container holding a list of callbacks.
    """

    callbacks: List[Callback] = field(default_factory=list)

    def append(self, callback):
        self.callbacks.append(callback)

    def set_params(self, params):
        for callback in self.callbacks:
            callback.set_params(params)

    def set_trainer(self, trainer):
        self.trainer = trainer
        for callback in self.callbacks:
            callback.set_trainer(trainer)

    def on_epoch_begin(self, epoch, logs=None):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch, logs)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_end(epoch, logs)

    def on_batch_begin(self, batch, logs=None):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_batch_begin(batch, logs)

    def on_batch_end(self, batch, logs=None):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_batch_end(batch, logs)

    def on_train_begin(self, logs=None):
        logs = logs or {}
        logs["start_time"] = time.time()
        for callback in self.callbacks:
            callback.on_train_begin(logs)

    def on_train_end(self, logs=None):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_end(logs)

@dataclass
class EarlyStopping(Callback):
    """EarlyStopping callback to exit the training loop if early_stopping_metric
    does not improve by a certain amount for a certain
    number of epochs.
    Parameters
    ---------
    early_stopping_metric : str
        Early stopping metric name
    is_maximize : bool
        Whether to maximize or not early_stopping_metric
    tol : float
        minimum change in monitored value to qualify as improvement.
        This number should be positive.
    patience : integer
        number of epochs to wait for improvement before terminating.
        the counter be reset after each improvement
    """

    early_stopping_metric: str
    is_maximize: bool
    tol: float = 0.0
    patience: int = 5
    logger: Any = None    

    def __post_init__(self):
        self.best_epoch = 0
        self.stopped_epoch = 0
        self.wait = 0
        self.best_weights = None
        self.best_loss = np.inf
        if self.is_maximize:
            self.best_loss = -self.best_loss
        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        current_loss = logs.get(self.early_stopping_metric)
        if current_loss is None:
            return

        loss_change = current_loss - self.best_loss
        max_improved = self.is_maximize and loss_change > self.tol
        min_improved = (not self.is_maximize) and (-loss_change > self.tol)
        if max_improved or min_improved:
            self.best_loss = current_loss
            self.best_epoch = epoch
            self.wait = 1
            self.best_weights = copy.deepcopy(self.trainer.network.state_dict())
        else:
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.trainer._stop_training = True
            self.wait += 1

    def on_train_end(self, logs=None):
        self.trainer.best_epoch = self.best_epoch
        self.trainer.best_cost = self.best_loss

        if self.best_weights is not None:
            self.trainer.network.load_state_dict(self.best_weights)

        if self.stopped_epoch > 0:
            msg = f"\nEarly stopping occurred at epoch {self.stopped_epoch}"
            msg += (
                f" with best_epoch = {self.best_epoch} and "
                + f"best_{self.early_stopping_metric} = {round(self.best_loss, 5)}"
            )
            if self.logger is not None:
                self.logger.info(msg)
        else:
            msg = (
                f"Stop training because you reached max_epochs = {self.trainer.max_epochs}"
                + f" with best_epoch = {self.best_epoch} and "
                + f"best_{self.early_stopping_metric} = {round(self.best_loss, 5)}"
            )
            if self.logger is not None:
                self.logger.info(msg)
        if self.logger is not None:
            self.logger.info("Best weights from best epoch are automatically used!")

@dataclass
class History(Callback):
    """Callback that records events into a `History` object.
    This callback is automatically applied to
    every SuperModule.
    Parameters
    ---------
    trainer : DeepRecoModel
        Model class to train
    verbose : int
        Print results every verbose iteration
    """

    trainer: Any
    verbose: int = 1
    logger: Any = None

    def __post_init__(self):
        super().__init__()
        self.samples_seen = 0.0
        self.total_time = 0.0

    def on_train_begin(self, logs=None):
        self.history = {}#{"loss": []}
        self.history.update({"lr": []})
        self.history.update({name: [] for name in self.trainer._metrics_names})
        self.start_time = logs["start_time"]
        self.epoch_loss = 0.0
        self.batch_logs = {}
        self.previous_batch = 0

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch_metrics = {}# {"loss": 0.0}
        self.samples_seen = 0.0

    def on_epoch_end(self, epoch, logs=None):
        #self.epoch_metrics["loss"] = self.epoch_loss
        for metric_name, metric_value in self.epoch_metrics.items():
            if metric_name not in self.history.keys():
                self.history[metric_name] = []
            self.history[metric_name].append(metric_value)
        for batch_log_name, batch_log_value in self.batch_logs.items():
            if batch_log_name not in self.history.keys():
                self.history[batch_log_name] = []
            self.history[batch_log_name].append(sum(batch_log_value) / len(batch_log_value))
        if self.verbose == 0:
            return
        if epoch % self.verbose != 0:
            return
        msg = f"epoch {epoch:<3}"
        for metric_name, metric_value in self.epoch_metrics.items():
            if metric_name != "lr":
                msg += f"| {metric_name:<3}: {np.round(metric_value, 5):<8}"
        self.total_time = int(time.time() - self.start_time)
        msg += f"|  {str(datetime.timedelta(seconds=self.total_time)) + 's':<6}"
        if self.logger is not None:
            self.logger.info(msg)

    def on_batch_end(self, batch, logs=None):
        batch_size = logs["batch_size"]
        
        # if epoch just started, clear batch logs
        if self.previous_batch > batch:
            self.batch_logs.clear()
        
        for key in logs:
            if key not in self.batch_logs:
                self.batch_logs[key] = list()
            self.batch_logs[key].append(logs[key])
        self.samples_seen += batch_size
        batch_count = int(len(self.trainer._dataset_train) / batch_size)
        self.previous_batch = batch
        if self.logger is not None:
            self.logger.info("Batch {} of {} batches trained. Batch size: {}".format(batch, batch_count, batch_size))

    def __getitem__(self, name):
        return self.history[name]

    def __repr__(self):
        return str(self.history)

    def __str__(self):
        return str(self.history)

@dataclass
class LRSchedulerCallback(Callback):
    """Wrapper for most torch scheduler functions.
    Parameters
    ---------
    scheduler_fn : torch.optim.lr_scheduler
        Torch scheduling class
    scheduler_params : dict
        Dictionnary containing all parameters for the scheduler_fn
    is_batch_level : bool (default = False)
        If set to False : lr updates will happen at every epoch
        If set to True : lr updates happen at every batch
        Set this to True for OneCycleLR for example
    """

    scheduler_fn: Any
    optimizer: Any
    scheduler_params: dict
    early_stopping_metric: str
    is_batch_level: bool = False

    def __post_init__(
        self,
    ):
        self.is_metric_related = hasattr(self.scheduler_fn, "is_better")
        self.scheduler = self.scheduler_fn(self.optimizer, **self.scheduler_params)
        super().__init__()

    def on_batch_end(self, batch, logs=None):
        if self.is_batch_level:
            self.scheduler.step()
        else:
            pass

    def on_epoch_end(self, epoch, logs=None):
        current_loss = logs.get(self.early_stopping_metric)
        if current_loss is None:
            return
        if self.is_batch_level:
            pass
        else:
            if self.is_metric_related:
                self.scheduler.step(current_loss)
            else:
                self.scheduler.step()
@dataclass
class TrainerSaverCallback(Callback):
    def __init__(self):
        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        return
        

@dataclass
class DataPlotterCallback(Callback):
    data_set: torch.utils.data.Dataset
    data_set_sample_count: int
    network_trainer: Any
    data_plotter: DataPlotter
    shuffle: bool = True
    logger: Any = None

    def __post_init__(self):
        super().__init__()
        self.epoch_logs = dict()
        self.epoch_logs_mins = dict()
        self.epoch_logs_maxs = dict()
        self.batch_logs = dict()
        self.batch_logs_min = 0
        self.batch_logs_max = 0
        self.inputs_outputs_real = []
        self.previous_batch = 0
        for sample_index in range(self.data_set_sample_count):
            if self.shuffle:
                sample_index = random.randint(0, len(self.data_set))
            input_real, output_real = self.data_set.__getitem__(sample_index)
            self.inputs_outputs_real.append((input_real, output_real))
    
    def on_batch_end(self, batch, logs=None):
        if self.previous_batch > batch:
            self.batch_logs.clear()

        for key in logs:
            if key not in self.batch_logs.keys():
                self.batch_logs[key] = list()
            self.batch_logs[key].append(logs[key])

        self.previous_batch = batch
        return

    def on_epoch_end(self, epoch, logs=None):
        inputs_real = []
        outputs_real = []
        outputs_fake = []
                
        dtype = self.trainer.network_dtype

        for input_real, output_real in self.inputs_outputs_real:
            inputs_real.append(input_real)
            outputs_real.append(output_real)

            input_real = self.trainer.to_tensor([input_real], dtype=dtype)
            
            output_fake = self.trainer.network(input_real)
            output_fake = output_fake[0]
            output_fake = self.trainer.to_numpy(output_fake)
            
            outputs_fake.append(output_fake)

        inputs_real = np.asarray(inputs_real)
        outputs_real = np.asarray(outputs_real)
        outputs_fake = np.asarray(outputs_fake)

        for key in logs:
            if key not in self.epoch_logs.keys():
                self.epoch_logs[key] = list()
            self.epoch_logs[key].append(logs[key])
        
        for key in self.batch_logs:
            if key not in self.epoch_logs.keys():
                self.epoch_logs[key] = list()
                self.epoch_logs_mins[key] = list()
                self.epoch_logs_maxs[key] = list()
            self.epoch_logs[key].append(sum(self.batch_logs[key])/len(self.batch_logs[key]))
            self.epoch_logs_mins[key].append(min(self.batch_logs[key]))
            self.epoch_logs_maxs[key].append(max(self.batch_logs[key]))

        # plot learning rate logs on same plot
        learning_rate_keys = [x for x in logs.keys() if 'lr_' in x or 'lr' in x]
        lr_logs = {}
        for lr_key in learning_rate_keys:
            lr_logs[lr_key] = self.epoch_logs[lr_key]
        self._plot_epoch_logs('learning_rates', lr_logs, self.trainer.output_directory_path, fill_to_zero=False)

        # plot network loss logs on same plot
        loss_keys = [x for x in self.batch_logs.keys() if 'loss' in x]
        self._plot_epoch_logs_windowed('losses', loss_keys, self.trainer.output_directory_path)

        # plot metric logs separately
        metric_keys = [x for x in logs.keys() if 'loss' not in x and 'lr_' not in x or 'lr' not in x]
        metric_logs = {}
        for metric_key in metric_keys:
            metric_logs[metric_key] = self.epoch_logs[metric_key]
            self._plot_epoch_log(metric_key, metric_logs[metric_key], self.trainer.output_directory_path)

        # plot qc images
        if self.logger is not None:
            self.logger.info("Plotting {} test samples".format(inputs_real.shape[0]))

        #utils.LOGGER.info("Plotting {} test samples".format(inputs_real.shape[0]))
        self.data_plotter.plot_batch(inputs_real, outputs_real, outputs_fake, epoch, self.trainer.output_directory_path)
    
    def _plot_epoch_logs_windowed(self, title, epoch_dictionary_keys, output_directory_path):
        if len(epoch_dictionary_keys) == 0 or epoch_dictionary_keys is None:
            return 
        
        for key in epoch_dictionary_keys:
            escape = False
            if key not in self.epoch_logs.keys():
                var_name = [ i for i, a in locals().iteritems() if a == self.epoch_logs][0]
                if self.logger is not None:
                    self.logger.info("Cannot plot filled plot because key: {} is not in dictionary: {}".format(key, var_name))
                escape = True
            if key not in self.epoch_logs_mins.keys():
                var_name = [ i for i, a in locals().iteritems() if a == self.epoch_logs][0]
                if self.logger is not None:
                    self.logger.info("Cannot plot filled plot because key: {} is not in dictionary: {}".format(key, var_name))
                escape = True
            if key not in self.epoch_logs_maxs.keys():
                var_name = [ i for i, a in locals().iteritems() if a == self.epoch_logs][0]
                if self.logger is not None:
                    self.logger.info("Cannot plot filled plot because key: {} is not in dictionary: {}".format(key, var_name))
                #utils.LOGGER.info("Cannot plot filled plot because key: {} is not in dictionary: {}".format(key, var_name))
                escape = True
            if escape == True:
                return

        first = self.epoch_logs[epoch_dictionary_keys[0]]
        epoch_x = np.arange(0, len(first))

        for key in epoch_dictionary_keys:
            epoch_y = self.epoch_logs[key]
            epoch_y_mins = self.epoch_logs_mins[key]
            epoch_y_maxs = self.epoch_logs_maxs[key]
            
            plt.plot(epoch_x, epoch_y, label=key)
            plt.fill_between(epoch_x, epoch_y_mins, epoch_y_maxs, alpha=0.2)
            plt.annotate('%0.5f' % epoch_y[len(epoch_y) - 1], xy=(1, epoch_y[len(epoch_y) - 1]), xytext=(8, 0), xycoords=('axes fraction', 'data'), textcoords='offset points')

            plt.legend(loc=4)
            plt.grid(True)
            plt.tight_layout()
            plt.xlabel('Epoch')
            plt.yscale('log')

        plt.savefig(os.path.join(output_directory_path, "{}.png".format(title)))
        plt.clf()
        plt.cla()
        plt.close()
        
        return 

    def _plot_epoch_logs(self, title, data_dictionary, output_directory_path, fill_to_zero:bool = True):
        first = data_dictionary[list(data_dictionary.keys())[0]]
        x = np.arange(0, len(first))

        for key in data_dictionary.keys():
            y = data_dictionary[key]
            plt.plot(x, y, label=key)

            if fill_to_zero == True:
                plt.fill_between(x, np.zeros(len(x)), y, alpha=0.2)

            plt.annotate('%0.5f' % y[len(y) - 1], xy=(1, y[len(y) - 1]), xytext=(8, 0), xycoords=('axes fraction', 'data'), textcoords='offset points')

            plt.legend(loc=4)
            plt.grid(True)
            plt.tight_layout()
            plt.xlabel('Epoch')

        plt.savefig(os.path.join(output_directory_path, "{}.png".format(title)))
        plt.clf()
        plt.cla()
        plt.close()
    
    def _plot_epoch_log(self, title, data, output_directory_path, fill_to_zero: bool=True):
        x = np.asarray(range(0, len(data)))
        y = data

        plt.figure(figsize=(20, 10))
        plt.plot(x, y, label=title, color='darkorange')

        if fill_to_zero == True:
            plt.fill_between(x, np.zeros(len(x)), y, alpha=0.2)
        #plt.ylim(-0.1, 2.1)
        plt.annotate('%0.5f' % y[len(y) - 1], xy=(1, y[len(y) - 1]), xytext=(8, 0), xycoords=('axes fraction', 'data'), textcoords='offset points')

        plt.legend(loc=4)
        plt.grid(True)
        plt.tight_layout()
        plt.xlabel('Epoch')

        plt.savefig(os.path.join(output_directory_path, "{}.png".format(title)))
        plt.clf()
        plt.cla()
        plt.close()


import os
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

import torch
import torch.nn

from nn_trainer.plotters import DataPlotter
matplotlib.use('Agg')

def make_path(output_path):
    if not os.path.isdir(output_path):
        os.makedirs(output_path)
    return output_path

class MnistCganImageDataPlotter(DataPlotter):
    def plot_batch(self, train_set_inputs, train_set_output, synthetic_output, epoch, output_dir):

        plt.clf()
        sample_dir = make_path(os.path.join(output_dir, str(epoch)))
        width = 0.4  # width of bar

        for idx, sample in enumerate(synthetic_output):
            output_path = os.path.join(sample_dir, "{}_noise".format(idx + 1))
            sample = sample[0]

            ax1 = plt.subplot(212)
            y_input = torch.nn.functional.one_hot(torch.tensor(train_set_inputs[idx,0]).long(), num_classes=10).numpy()
            x_input = np.arange(len(y_input))
            #y_input = train_set_inputs[idx,:]
            ax1.set_xticks(x_input)

            plt.bar(x_input, y_input, label='Input_{}'.format(0), color='indigo')
            plt.ylabel('Input {}'.format(0))
            plt.ylim(0, 1.1)
            plt.legend(loc=4)
            plt.grid(axis='x')

            ax2 = plt.subplot(221)
            plt.imshow(train_set_output[idx,0,:,:])
            plt.colorbar()

            ax3 = plt.subplot(222)
            plt.imshow(synthetic_output[idx,0,:,:])
            plt.colorbar()
            
            plt.savefig(output_path + '.png')
            plt.clf()
            plt.cla()
            plt.close()
        pass


from .data import *
from .models import *
from .plotters import *
from .trainers import *

from .metrics import *
from .callbacks import *

from .utils import parse_arguments, make_path

VERSION = "0.0.2.6"



class DataPlotter(object):
    def plot_batch(self, train_set_inputs, train_set_output, synthetic_output, epoch, output_dir):
        raise NotImplementedError("plot_batch method must be implemented")


from abc import abstractmethod
import io
import os
import os.path

import json
from pathlib import Path
import shutil
import zipfile
import warnings
import copy
import numpy
from numpy import ndarray
from nn_trainer.models.networks.networks_1d import init_weights

import datetime

from nn_trainer.metrics import *
from nn_trainer.callbacks import *

import torch
import torch.nn
import torch.optim
import torch.utils.data as td
import torch.nn.functional as f

from typing import Dict

class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.int64):
            return int(obj)
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, torch.Tensor):
            return obj.cpu().numpy()
        elif isinstance(obj, torch.dtype):
            return str(obj)

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)

def make_path(output_path):
    if not os.path.isdir(output_path):
        os.makedirs(output_path)
    return output_path

class AbstractNnTrainer(object):
    """
    Base class for Neural Network trainers
    """
    def __init__(
        self,
        args,
        neural_network: torch.nn.Module, 
        optimizer_fn: Any = torch.optim.Adam,
        optimizer_params: Dict = dict(lr=2e-2),
        scheduler_fn: Any = None,
        scheduler_params: Dict = field(default_factory=dict),
        dtype: Any = torch.float32,
        verbose: bool = False,
        patience: int = 0,
        logger = None
        ):
        
        """
        The base constructor

        Args:
            args (Any): Command line args
            neural_network (torch.nn.Module): The network intended to be trained.  If implementation is a gan, then this refers to the 'generator' network
            optimizer_fn (Any, optional): factory method for the optimizer to be constructed and used for training during initialization. Defaults to torch.optim.Adam.
            optimizer_params (Dict, optional): parameters for the optimizer function. Defaults to dict(lr=2e-2).
            scheduler_fn (Any, optional): factory method for the optimizer scheduler to be constructed and used for training during initialization. Defaults to None.
            scheduler_params (Dict, optional): parameters for the optimizer scheduler function. Defaults to field(default_factory=dict).
            dtype (Any, optional): the type of data to allocate memory with during training. Defaults to torch.float32.
            verbose (bool, optional): Will print as much output as possible if True. Defaults to False.
            patience (int, optional): describes the stopping criteria number of epochs over which training losses are relatively the same. Defaults to 0.
            logger ([type], optional): logging utility to be used during runtime. must have an info method. Defaults to None.
        """
        self._args = args
        self._verbose = verbose
        self._patience = patience
        self._logger = logger
        
        self._is_cuda_enabled = True if torch.cuda.is_available() else False
        self._device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self._ngpus = torch.cuda.device_count() if torch.cuda.is_available() else 0
        self._dtype = dtype

        self._net_g = neural_network.type(dtype)
        self._net_g.to(self._device)
        self._net_g_optimizer_fn = optimizer_fn
        self._net_g_optimizer_params = optimizer_params
        self._net_g_scheduler_fn = scheduler_fn
        self._net_g_scheduler_params = scheduler_params

        self._net_g_optimizer = self._net_g_optimizer_fn(self._net_g.parameters(), **self._net_g_optimizer_params)

        self._epoch_count = args['max_epoch_count']
        self._batch_size = args['batch_size']
        self._model_directory_path = make_path(os.path.join(args['output_dir'], "{}_{}".format(datetime.datetime.now().strftime("%Y%m%d%H%M%S"), self.__class__.__name__))) 
        self._generator_model_directory_path = make_path(os.path.join(self._model_directory_path, self._net_g.__class__.__name__))

    @property
    def device(self):
        """
        The device on which memory is allocated and processing is performed during training

        Returns:
            torch.device: the torch device training is performed on, either 'cpu' or 'cuda:0'
        """
        return self._device

    @property
    def network(self):
        """
        The neural network indended to be trained

        Returns:
            [torch.nn.Module]: the neural network indended to be trained
        """
        return self._net_g

    @property
    def network_dtype(self):
        """
        The data type used for memory allocation in the network (mixed dtypes may be accommodated in the future)

        Returns:
            torch.dtype: the type of the data used for memory allocation (ex. torch.float32)
        """
        return self._dtype

    @property
    def output_directory_path(self):
        """
        The default output directory path for model persistence (to be removed in future instances in lieu of using appropriate callbacks)

        Returns:
            str: directory path of output data
        """
        return self._model_directory_path
    
    def to_tensor(self, data:Any, dtype=torch.float32):
        """
        A utility method for converting data to tensor objects 

        Args:
            data (Any): The data to convert to a tensor
            dtype (torch.dtype, optional): Defaults to torch.float32.

        Raises:
            TypeError: If data is not a list, or not convertible to a numpy ndarray.

        Returns:
            torch.Tensor: A tensor
        """
        if isinstance(data, list):
            result = torch.tensor(numpy.asarray(data), dtype=dtype).to(device=self._device)
        elif isinstance(data, ndarray):
            result = torch.tensor(data, dtype=dtype).to(device=self._device)
        elif isinstance(data, torch.Tensor):
            result = data.clone().detach().requires_grad_(data.requires_grad).to(device=self._device, dtype=self._dtype)
        else:
            raise TypeError("type {} of data is not acceptable".format(type(data)))
        return result

    def to_numpy(self, data: torch.Tensor):
        """
        Converts a torch.Tensor to a numpy.ndarray

        Args:
            data (torch.Tensor): The array to convert

        Returns:
            numpy.ndarray: The converted array
        """
        if data.requires_grad:
            data = data.detach()
        if self._device == torch.device("cuda:0"):
            data = data.cpu()
        return data.numpy()

    @abstractmethod
    def train(
        self,
        training_data_set: td.Dataset,
        validation_data_set: td.Dataset, 
        callbacks: List[Callback] = [], 
        loss_fn = torch.nn.MSELoss(),
        metrics: List[Metric] = [],
        ):
        """
        The training method
        
        Args:
            training_data_set (td.Dataset): The dataset intended to be used for training
            validation_data_set (td.Dataset): The dataset intended to be used for validation. all metrics defined will be invokd on this set
            callbacks (List[Callback], optional): The set of callbacks receiving event based data. Defaults to [].
            loss_fn ([type], optional): The loss function to be used as the objective function during training. Defaults to torch.nn.MSELoss().
            metrics (List[Metric], optional): List of metrics to be invoked on validation data. Defaults to [].

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError("train method must be implemented")

    def save_model(self, model_name):
        """Saving the model.  This method reuses the model output directory property 
        initialized during construction of this object.
        Parameters
        ----------
        model_name : str
            The name to give to the model. Recommendation: use epoch in the name
        Returns
        -------
        str
            input filepath with ".pt" appended
        """
        path = self.output_directory_path + "\\{}".format(model_name)

        saved_params = {}
        init_params = {}
        for key, val in self._net_g.state_dict().items():
            if isinstance(val, type):
                # Don't save torch specific params
                continue
            else:
                init_params[key] = { 'shape': val.shape, 'dtype': val.dtype }
        init_params["device_type"] = self._device.type
        init_params["device_index"] = self._device.index
        saved_params["init_params"] = init_params

        # Create folder
        Path(path).mkdir(parents=True, exist_ok=True)
        
        # Save models params
        with open(Path(path).joinpath("model_params.json"), "w", encoding="utf8") as f:
            json.dump(saved_params, f, cls=ComplexEncoder, indent=4)

        # Save state_dict
        torch.save(self.network.state_dict(), Path(path).joinpath("network.pt"))
        shutil.make_archive(path, "zip", path)
        shutil.rmtree(path)
        print(f"Successfully saved model at {path}.zip")
        return f"{path}.zip"

    def load_model(self, filepath):
        """Load TabNet model.
        Parameters
        ----------
        filepath : str
            Path of the model.
        """
        path = Path(filepath)
        if path.suffix == "":
            path = Path(filepath + ".zip")
        
        if path.suffix != ".zip":
            raise TypeError("non zip files are not supported")

        try:
            with zipfile.ZipFile(path) as z:
                with z.open("model_params.json") as f:
                    loaded_params = json.load(f)
                    loaded_params["init_params"]["device_name"] = self._device
                with z.open("network.pt") as f:
                    try:
                        saved_state_dict = torch.load(f, map_location=self.device)
                    except io.UnsupportedOperation:
                        # In Python <3.7, the returned file object is not seekable (which at least
                        # some versions of PyTorch require) - so we'll try buffering it in to a
                        # BytesIO instead:
                        saved_state_dict = torch.load(io.BytesIO(f.read()), map_location=self.device)
        except KeyError:
            raise KeyError("Your zip file is missing at least one component")

        #self.__init__(**loaded_params["init_params"])

        #self._set_network()
        self.network.load_state_dict(saved_state_dict)
        self.network.eval()
        #self.load_class_attrs(loaded_params["class_attrs"])

        return

    @abstractmethod
    def _set_callbacks(self, custom_callbacks):
        raise NotImplementedError("_set_callbacks method must be implemented")
    
    def _set_metrics(self, metrics, eval_names):
        """Set attributes relative to the metrics.
        Parameters
        ----------
        metrics : list of str
            List of eval metric names.
        eval_names : list of str
            List of eval set names.
        """
        metrics = metrics or [RMSE]

        metrics = check_metrics(metrics)
        # Set metric container for each sets
        self._metric_container_dict = {}
        for name in eval_names:
            self._metric_container_dict.update({name: MetricContainer(metrics, prefix=f"{name}_metric__")})

        self._metrics = []
        self._metrics_names = []
        for _, metric_container in self._metric_container_dict.items():
            self._metrics.extend(metric_container.metrics)
            self._metrics_names.extend(metric_container.names)

        # Early stopping metric is the last eval metric
        self.early_stopping_metric = (self._metrics_names[-1] if len(self._metrics_names) > 0 else None)

    def _predict_epoch(self, name, data):
        """
        Predict an epoch and update metrics.

        Parameters
        ----------
        name : str
            Name of the validation set
        loader : torch.utils.data.Dataloader
                DataLoader with validation set
        """
        # Setting network on evaluation mode
        self._net_g.eval()

        list_y_true = []
        list_y_score = []
        
        # Main loop
        X = data[0]
        y = data[1]
        scores = self._predict_batch(X)
        list_y_true.append(y)
        list_y_score.append(scores)

        y_true, scores = self._stack_batches(list_y_true, list_y_score)

        metrics_logs = self._metric_container_dict[name](y_true, scores)
        self._net_g.train()
        self.history.epoch_metrics.update(metrics_logs)
        return

    def _predict_batch(self, X):
        """
        Predict one batch of data.

        Parameters
        ----------
        X : torch.Tensor
            Owned products

        Returns
        -------
        np.array
            model scores
        """
        X = self.to_tensor(X)
        #X = X.to(self._device).float()

        # compute model output
        scores = self._net_g(X)

        if isinstance(scores, list):
            scores = [self.to_numpy(x) for x in scores]
        else:
            scores = self.to_numpy(scores)#cpu().detach().numpy()

        return scores
    
    def _stack_batches(self, list_y_true, list_y_score):
        y_true = np.vstack(list_y_true)
        y_score = np.vstack(list_y_score)
        return y_true, y_score


import datetime
import logging
import os, time
import errno
import torch

def make_path(output_path):
    if not os.path.isdir(output_path):
        os.makedirs(output_path)
    return output_path

class Logger(object):
    def __init__(self, reuse:bool = True, log_name:str = 'nn_logger'):
        if reuse:
            self._logger = logging.getLogger(log_name)
        else:
            self._model_directory_path = make_path("{}_{}".format(datetime.datetime.now().strftime("%Y%m%d%H%M%S"), self.__class__.__name__)) 
            self._logger = logging.getLogger()
        return
    
    def log(self, d_error, g_error, epoch_index, batch_index):
         # var_class = torch.autograd.variable.Variable
        if isinstance(d_error, torch.autograd.Variable):            d_error = d_error.data.cpu().numpy()
        if isinstance(g_error, torch.autograd.Variable):            g_error = g_error.data.cpu().numpy()

        step = Logger._step(epoch_index, batch_index, self.num_batches)
        self.steps.append(step)
        self.d_losses.append(d_error)
        self.g_losses.append(g_error)

    def display_status(self, d_error, g_error, d_pred_real, d_pred_fake, epoch_index, batch_index):
        if isinstance(d_error, torch.autograd.Variable):
            d_error = d_error.data.cpu().numpy()
        if isinstance(g_error, torch.autograd.Variable):
            g_error = g_error.data.cpu().numpy()

        current_elapsed_time = time.time() - self.object_construction_time
        print('Elapsed Time: {:.4f} Epoch: [{}/{}], Batch Num: [{}/{}] Discriminator Loss: {:.4f}, Generator Loss: {:.4f} D(x,y): {:.4f}, D(G(z,y),y): {:.4f}'.format(current_elapsed_time, epoch_index, self.num_epochs, batch_index, self.num_batches, d_error, g_error, d_pred_real.mean(), d_pred_fake.mean()))
        print('     Discriminator Loss: {:.4f}, Generator Loss: {:.4f} D(x,y): {:.4f}, D(G(z,y),y): {:.4f}'.format(d_error, g_error, d_pred_real.mean(), d_pred_fake.mean()))
        print('     D(x,y): {:.4f}, D(G(z,y),y): {:.4f}'.format(d_pred_real.mean(), d_pred_fake.mean()))

        return
    
    @staticmethod
    def _step(epoch, n_batch, num_batches):
        return epoch * num_batches + n_batch

    @staticmethod
    def _make_dir(directory):
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        pass